/**
 * This file is part of the "libterminal" project
 *   Copyright (c) 2019-2020 Christian Parpart <christian@parpart.family>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once

#include <vtbackend/InputGenerator.h>
#include <vtbackend/InputHandler.h>
#include <vtbackend/RenderBuffer.h>
#include <vtbackend/ScreenEvents.h>
#include <vtbackend/Selector.h>
#include <vtbackend/Sequence.h>
#include <vtbackend/Settings.h>
#include <vtbackend/TerminalState.h>
#include <vtbackend/ViInputHandler.h>
#include <vtbackend/Viewport.h>
#include <vtbackend/cell/CellConcept.h>
#include <vtbackend/cell/CellConfig.h>
#include <vtbackend/primitives.h>

#include <vtpty/Pty.h>

#include <crispy/defines.h>

#include <fmt/format.h>

#include <atomic>
#include <chrono>
#include <functional>
#include <memory>
#include <mutex>
#include <string_view>
#include <type_traits>
#include <vector>

namespace terminal
{

template <typename Cell>
CRISPY_REQUIRES(CellConcept<Cell>)
class Screen;

/// Helping information to visualize IME text that has not been comitted yet.
struct InputMethodData
{
    // If this string is non-empty, the IME is active and the given data
    // shall be displayed at the cursor's location.
    std::string preeditString;
};

/// Terminal API to manage input and output devices of a pseudo terminal, such as keyboard, mouse, and screen.
///
/// With a terminal being attached to a Process, the terminal's screen
/// gets updated according to the process' outputted text,
/// whereas input to the process can be send high-level via the various
/// send(...) member functions.
class Terminal
{
  public:
    class Events
    {
      public:
        virtual ~Events() = default;

        virtual void requestCaptureBuffer(LineCount /*lines*/, bool /*logical*/) {}
        virtual void bell() {}
        virtual void bufferChanged(ScreenType) {}
        virtual void renderBufferUpdated() {}
        virtual void screenUpdated() {}
        virtual FontDef getFontDef() { return {}; }
        virtual void setFontDef(FontDef const& /*_fontSpec*/) {}
        virtual void copyToClipboard(std::string_view /*_data*/) {}
        virtual void inspect() {}
        virtual void notify(std::string_view /*_title*/, std::string_view /*_body*/) {}
        virtual void onClosed() {}
        virtual void pasteFromClipboard(unsigned /*count*/, bool /*strip*/) {}
        virtual void onSelectionCompleted() {}
        virtual void requestWindowResize(LineCount, ColumnCount) {}
        virtual void requestWindowResize(Width, Height) {}
        virtual void setWindowTitle(std::string_view /*_title*/) {}
        virtual void setTerminalProfile(std::string const& /*_configProfileName*/) {}
        virtual void discardImage(Image const&) {}
        virtual void inputModeChanged(ViMode /*mode*/) {}
        virtual void updateHighlights() {}
        virtual void playSound(Sequence::Parameters const&) {}
        virtual void cursorPositionChanged() {}
        virtual void onScrollOffsetChanged(ScrollOffset) {}
    };

    Terminal(Events& _eventListener,
             std::unique_ptr<Pty> _pty,
             Settings _factorySettings,
             std::chrono::steady_clock::time_point _now /* = std::chrono::steady_clock::now()*/);
    ~Terminal() = default;

    void start();

    void setRefreshRate(RefreshRate _refreshRate);
    void setLastMarkRangeOffset(LineOffset _value) noexcept;

    void setMaxHistoryLineCount(MaxHistoryLineCount _maxHistoryLineCount);
    LineCount maxHistoryLineCount() const noexcept;

    void setTerminalId(VTType _id) noexcept { state_.terminalId = _id; }

    void setMaxImageSize(ImageSize size) noexcept { state_.effectiveImageCanvasSize = size; }

    void setMaxImageSize(ImageSize _effective, ImageSize _limit)
    {
        state_.effectiveImageCanvasSize = _effective;
        settings_.maxImageSize = _limit;
    }

    bool isModeEnabled(AnsiMode m) const noexcept { return state_.modes.enabled(m); }
    bool isModeEnabled(DECMode m) const noexcept { return state_.modes.enabled(m); }
    void setMode(AnsiMode _mode, bool _enable);
    void setMode(DECMode _mode, bool _enable);

    void setTopBottomMargin(std::optional<LineOffset> _top, std::optional<LineOffset> _bottom);
    void setLeftRightMargin(std::optional<ColumnOffset> _left, std::optional<ColumnOffset> _right);

    void moveCursorTo(LineOffset _line, ColumnOffset _column);
    void saveCursor();
    void restoreCursor();
    void restoreCursor(Cursor const& _savedCursor);

    void setGraphicsRendition(GraphicsRendition _rendition);
    void setForegroundColor(Color _color);
    void setBackgroundColor(Color _color);
    void setUnderlineColor(Color _color);
    void setHighlightRange(HighlightRange _range);

    // {{{ cursor
    [[nodiscard]] Cursor const& cursor() const noexcept { return state_.cursor; }
    [[nodiscard]] constexpr CellLocation realCursorPosition() const noexcept
    {
        return state_.cursor.position;
    }

    /// Clamps given coordinates, respecting DECOM (Origin Mode).
    [[nodiscard]] CellLocation clampCoordinate(CellLocation coord) const noexcept
    {
        if (state_.cursor.originMode)
            return clampToOrigin(coord);
        else
            return clampToScreen(coord);
    }

    /// Clamps given logical coordinates to margins as used in when DECOM (origin mode) is enabled.
    [[nodiscard]] CellLocation clampToOrigin(CellLocation coord) const noexcept
    {
        return { std::clamp(coord.line, LineOffset { 0 }, currentScreen().margin().vertical.to),
                 std::clamp(coord.column, ColumnOffset { 0 }, currentScreen().margin().horizontal.to) };
    }

    [[nodiscard]] LineOffset clampedLine(LineOffset _line) const noexcept
    {
        return std::clamp(_line, LineOffset(0), boxed_cast<LineOffset>(settings_.pageSize.lines) - 1);
    }

    [[nodiscard]] ColumnOffset clampedColumn(ColumnOffset _column) const noexcept
    {
        return std::clamp(_column, ColumnOffset(0), boxed_cast<ColumnOffset>(settings_.pageSize.columns) - 1);
    }

    [[nodiscard]] CellLocation clampToScreen(CellLocation coord) const noexcept
    {
        return { clampedLine(coord.line), clampedColumn(coord.column) };
    }

    // Tests if given coordinate is within the visible screen area.
    [[nodiscard]] constexpr bool contains(CellLocation _coord) const noexcept
    {
        return LineOffset(0) <= _coord.line && _coord.line < boxed_cast<LineOffset>(settings_.pageSize.lines)
               && ColumnOffset(0) <= _coord.column
               && _coord.column <= boxed_cast<ColumnOffset>(settings_.pageSize.columns);
    }

    [[nodiscard]] bool isCursorInViewport() const noexcept
    {
        return viewport().isLineVisible(cursor().position.line);
    }
    // }}}

    [[nodiscard]] constexpr ImageSize cellPixelSize() const noexcept { return state_.cellPixelSize; }
    constexpr void setCellPixelSize(ImageSize _cellPixelSize) { state_.cellPixelSize = _cellPixelSize; }

    /// Retrieves the time point this terminal instance has been spawned.
    [[nodiscard]] std::chrono::steady_clock::time_point currentTime() const noexcept { return currentTime_; }

    /// Retrieves reference to the underlying PTY device.
    [[nodiscard]] Pty& device() noexcept { return *pty_; }
    [[nodiscard]] Pty const& device() const noexcept { return *pty_; }

    [[nodiscard]] PageSize pageSize() const noexcept { return pty_->pageSize(); }

    [[nodiscard]] PageSize totalPageSize() const noexcept { return settings_.pageSize; }

    /// Resizes the terminal screen to the given amount of grid cells with their pixel dimensions.
    /// Important! In case a status line is currently visible, the status line count is being
    /// accumulated into the screen size, too.
    void resizeScreen(PageSize _cells, std::optional<ImageSize> _pixels = std::nullopt);
    void resizeScreenInternal(PageSize _cells, std::optional<ImageSize> _pixels);

    /// Implements semantics for  DECCOLM / DECSCPP.
    void resizeColumns(ColumnCount _newColumnCount, bool _clear);

    void clearScreen();

    void setMouseProtocolBypassModifier(Modifier value) { settings_.mouseProtocolBypassModifier = value; }
    void setMouseBlockSelectionModifier(Modifier value) { settings_.mouseBlockSelectionModifier = value; }

    bool isMouseGrabbedByApp() const noexcept;

    // {{{ input proxy
    using Timestamp = std::chrono::steady_clock::time_point;
    bool sendKeyPressEvent(Key _key, Modifier _modifier, Timestamp _now);
    bool sendCharPressEvent(char32_t _char, Modifier _modifier, Timestamp _now);
    bool sendMousePressEvent(Modifier _modifier,
                             MouseButton _button,
                             PixelCoordinate _pixelPosition,
                             bool _uiHandledHint,
                             Timestamp _now);
    bool sendMouseMoveEvent(Modifier _modifier,
                            CellLocation _pos,
                            PixelCoordinate _pixelPosition,
                            bool _uiHandledHint,
                            Timestamp _now);
    bool sendMouseReleaseEvent(Modifier _modifier,
                               MouseButton _button,
                               PixelCoordinate _pixelPosition,
                               bool _uiHandledHint,
                               Timestamp _now);
    bool sendFocusInEvent();
    bool sendFocusOutEvent();
    void sendPaste(std::string_view _text); // Sends verbatim text in bracketed mode to application.
    void sendPasteFromClipboard(unsigned count, bool strip)
    {
        eventListener_.pasteFromClipboard(count, strip);
    }

    bool handleMouseSelection(Modifier _modifier, Timestamp _now);

    void inputModeChanged(ViMode mode) { eventListener_.inputModeChanged(mode); }
    void updateHighlights() { eventListener_.updateHighlights(); }
    void playSound(terminal::Sequence::Parameters const& params_) { eventListener_.playSound(params_); }

    bool applicationCursorKeys() const noexcept { return state_.inputGenerator.applicationCursorKeys(); }
    bool applicationKeypad() const noexcept { return state_.inputGenerator.applicationKeypad(); }

    bool hasInput() const noexcept;
    size_t pendingInputBytes() const noexcept;
    void flushInput();

    std::string_view peekInput() const noexcept { return state_.inputGenerator.peek(); }
    // }}}

    /// Writes a given VT-sequence to screen.
    void writeToScreen(std::string_view _text);

    /// Writes a given VT-sequence to screen - but without acquiring the lock (must be already acquired).
    void writeToScreenInternal(std::string_view _text);

    // viewport management
    [[nodiscard]] Viewport& viewport() noexcept { return viewport_; }
    [[nodiscard]] Viewport const& viewport() const noexcept { return viewport_; }

    // {{{ Screen Render Proxy
    std::optional<std::chrono::milliseconds> nextRender() const;

    uint64_t tick(std::chrono::steady_clock::time_point _now)
    {
        auto const changes = changes_.exchange(0);
        currentTime_ = _now;
        updateCursorVisibilityState();
        if (isBlinkOnScreen())
        {
            tie(_rapidBlinker.state, _lastRapidBlink) = nextBlinkState(_rapidBlinker, _lastRapidBlink);
            tie(_slowBlinker.state, _lastBlink) = nextBlinkState(_slowBlinker, _lastBlink);
        }
        return changes;
    }
    // }}}

    // {{{ RenderBuffer synchronization API

    /// Ensures the terminals event loop is interrupted
    /// and the render buffer is refreshed.
    ///
    void breakLoopAndRefreshRenderBuffer();

    /// Refreshes the render buffer.
    /// When this function returns, the back buffer is updated
    /// and it is attempted to swap the back/front buffers.
    /// but the swap has NOT been invoked yet.
    ///
    /// @param _locked whether or not the Terminal object's lock is already held by the caller.
    ///
    /// @retval true   front buffer now contains the refreshed render buffer.
    /// @retval false  back buffer contains the refreshed render buffer,
    ///                and RenderDoubleBuffer::swapBuffers() must again
    ///                be successfully invoked to swap back/front buffers
    ///                in order to access the refreshed render buffer.
    ///
    /// @note The current time must have been updated in order to get the
    ///       correct cursor blinking state drawn.
    ///
    /// @see RenderDoubleBuffer::swapBuffers()
    /// @see renderBuffer()
    ///
    bool refreshRenderBuffer(bool _locked = false);

    /// Eventually refreshes the render buffer iff
    /// - the screen contents has changed AND refresh rate satisfied,
    /// - viewport has changed, or
    /// - refreshing the render buffer was explicitly requested.
    ///
    /// @param _now    the current time
    /// @param _locked whether or not the Terminal object's lock is already held by the caller.
    ///
    /// @see RenderDoubleBuffer::swapBuffers()
    /// @see renderBuffer()
    bool ensureFreshRenderBuffer(bool _locked = false);

    /// Aquuires read-access handle to front render buffer.
    ///
    /// This also acquires the reader lock and releases it automatically
    /// upon RenderBufferRef destruction.
    ///
    /// @see ensureFreshRenderBuffer()
    /// @see refreshRenderBuffer()
    [[nodiscard]] RenderBufferRef renderBuffer() const { return renderBuffer_.frontBuffer(); }

    [[nodiscard]] RenderBufferState renderBufferState() const noexcept { return renderBuffer_.state; }

    /// Updates the IME preedit-string to be rendered when IME is composing a new input.
    /// Passing an empty string effectively disables IME rendering.
    void updateInputMethodPreeditString(std::string preeditString);
    // }}}

    void lock() const
    {
        outerLock_.lock();
        innerLock_.lock();
    }

    void unlock() const
    {
        outerLock_.unlock();
        innerLock_.unlock();
    }

    [[nodiscard]] ColorPalette const& colorPalette() const noexcept { return state_.colorPalette; }
    [[nodiscard]] ColorPalette& colorPalette() noexcept { return state_.colorPalette; }
    [[nodiscard]] ColorPalette& defaultColorPalette() noexcept { return state_.defaultColorPalette; }

    void pushColorPalette(size_t slot);
    void popColorPalette(size_t slot);
    void reportColorPaletteStack();

    [[nodiscard]] ScreenBase& currentScreen() noexcept { return currentScreen_.get(); }
    [[nodiscard]] ScreenBase const& currentScreen() const noexcept { return currentScreen_.get(); }

    [[nodiscard]] ScreenBase& activeDisplay() noexcept
    {
        switch (state_.activeStatusDisplay)
        {
            case ActiveStatusDisplay::Main: return currentScreen_.get();
            case ActiveStatusDisplay::StatusLine: return hostWritableStatusLineScreen_;
        }
        crispy::unreachable();
    }

    bool isPrimaryScreen() const noexcept { return state_.screenType == ScreenType::Primary; }
    bool isAlternateScreen() const noexcept { return state_.screenType == ScreenType::Alternate; }
    ScreenType screenType() const noexcept { return state_.screenType; }
    void setScreen(ScreenType screenType);

    ScreenBase& screenForType(ScreenType type) noexcept
    {
        switch (type)
        {
            case ScreenType::Primary: return primaryScreen_;
            case ScreenType::Alternate: return alternateScreen_;
        }
        crispy::unreachable();
    }

    void setHighlightTimeout(std::chrono::milliseconds timeout) noexcept
    {
        settings_.highlightTimeout = timeout;
    }

    // clang-format off
    [[nodiscard]] Screen<PrimaryScreenCell> const& primaryScreen() const noexcept { return primaryScreen_; }
    [[nodiscard]] Screen<PrimaryScreenCell>& primaryScreen() noexcept { return primaryScreen_; }
    [[nodiscard]] Screen<AlternateScreenCell> const& alternateScreen() const noexcept { return alternateScreen_; }
    [[nodiscard]] Screen<AlternateScreenCell>& alternateScreen() noexcept { return alternateScreen_; }
    [[nodiscard]] Screen<StatusDisplayCell> const& hostWritableStatusLineDisplay() const noexcept { return hostWritableStatusLineScreen_; }
    [[nodiscard]] Screen<StatusDisplayCell> const& indicatorStatusLineDisplay() const noexcept { return indicatorStatusScreen_; }
    // clang-format on

    [[nodiscard]] bool isLineWrapped(LineOffset _lineNumber) const noexcept
    {
        return isPrimaryScreen() && primaryScreen_.isLineWrapped(_lineNumber);
    }

    [[nodiscard]] CellLocation currentMousePosition() const noexcept { return currentMousePosition_; }

    [[nodiscard]] std::optional<CellLocation> currentMouseGridPosition() const noexcept
    {
        if (currentScreen_.get().contains(currentMousePosition_))
            return viewport_.translateScreenToGridCoordinate(currentMousePosition_);
        return std::nullopt;
    }

    // {{{ cursor management
    CursorDisplay cursorDisplay() const noexcept { return settings_.cursorDisplay; }
    void setCursorDisplay(CursorDisplay _value);

    CursorShape cursorShape() const noexcept { return settings_.cursorShape; }
    void setCursorShape(CursorShape _value);

    bool cursorBlinkActive() const noexcept { return cursorBlinkState_; }

    bool cursorCurrentlyVisible() const noexcept
    {
        return isModeEnabled(DECMode::VisibleCursor)
               && (cursorDisplay() == CursorDisplay::Steady || cursorBlinkState_);
    }

    bool isBlinkOnScreen() const noexcept { return _lastRenderPassHints.containsBlinkingCells; }

    std::chrono::steady_clock::time_point lastCursorBlink() const noexcept { return lastCursorBlink_; }

    constexpr void setCursorBlinkingInterval(std::chrono::milliseconds _value)
    {
        settings_.cursorBlinkInterval = _value;
    }

    constexpr std::chrono::milliseconds cursorBlinkInterval() const noexcept
    {
        return settings_.cursorBlinkInterval;
    }
    // }}}

    // {{{ selection management
    // TODO: move you, too?
    void setWordDelimiters(std::string const& _wordDelimiters);
    std::u32string const& wordDelimiters() const noexcept { return settings_.wordDelimiters; }

    Selection const* selector() const noexcept { return selection_.get(); }
    Selection* selector() noexcept { return selection_.get(); }
    std::chrono::milliseconds highlightTimeout() const noexcept { return settings_.highlightTimeout; }

    template <typename RenderTarget>
    void renderSelection(RenderTarget _renderTarget) const
    {
        if (!selection_)
            return;

        if (isPrimaryScreen())
            terminal::renderSelection(
                *selection_, [&](CellLocation _pos) { _renderTarget(_pos, primaryScreen_.at(_pos)); });
        else
            terminal::renderSelection(
                *selection_, [&](CellLocation _pos) { _renderTarget(_pos, alternateScreen_.at(_pos)); });
    }

    void clearSelection();

    /// Tests whether some area has been selected.
    bool isSelectionAvailable() const noexcept
    {
        return selection_ && selection_->state() != Selection::State::Waiting;
    }
    bool isSelectionInProgress() const noexcept
    {
        return selection_ && selection_->state() != Selection::State::Complete;
    }
    bool isSelectionComplete() const noexcept
    {
        return selection_ && selection_->state() == Selection::State::Complete;
    }

    /// Tests whether given absolute coordinate is covered by a current selection.
    bool isSelected(CellLocation _coord) const noexcept
    {
        return selection_ && selection_->state() != Selection::State::Waiting && selection_->contains(_coord);
    }

    /// Tests whether given line offset is intersecting with selection.
    bool isSelected(LineOffset line) const noexcept
    {
        return selection_ && selection_->state() != Selection::State::Waiting
               && selection_->containsLine(line);
    }

    bool isHighlighted(CellLocation _cell) const noexcept;
    bool blinkState() const noexcept { return _slowBlinker.state; }
    bool rapidBlinkState() const noexcept { return _rapidBlinker.state; }

    /// Sets or resets to a new selection.
    void setSelector(std::unique_ptr<Selection> _selector) { selection_ = std::move(_selector); }

    /// Tests whether or not some grid cells are selected.
    bool selectionAvailable() const noexcept { return !!selection_; }

    bool visualizeSelectedWord() const noexcept { return settings_.visualizeSelectedWord; }
    void setVisualizeSelectedWord(bool enabled) noexcept { settings_.visualizeSelectedWord = enabled; }
    // }}}

    [[nodiscard]] std::string extractSelectionText() const;
    [[nodiscard]] std::string extractLastMarkRange() const;

    /// Tests whether or not the mouse is currently hovering a hyperlink.
    [[nodiscard]] bool isMouseHoveringHyperlink() const noexcept { return hoveringHyperlink_.load(); }

    /// Retrieves the HyperlinkInfo that is currently behing hovered by the mouse, if so,
    /// or a nothing otherwise.
    [[nodiscard]] std::shared_ptr<HyperlinkInfo const> tryGetHoveringHyperlink() const noexcept
    {
        if (auto const gridPosition = currentMouseGridPosition())
            return currentScreen_.get().hyperlinkAt(*gridPosition);
        return {};
    }

    bool processInputOnce();

    void markScreenDirty() { screenDirty_ = true; }
    [[nodiscard]] bool screenDirty() const noexcept { return screenDirty_; }

    [[nodiscard]] uint64_t lastFrameID() const noexcept { return lastFrameID_.load(); }

    // Screen's EventListener implementation
    //
    void requestCaptureBuffer(LineCount lines, bool logical);
    void bell();
    void bufferChanged(ScreenType);
    void scrollbackBufferCleared();
    void screenUpdated();
    [[nodiscard]] FontDef getFontDef();
    void setFontDef(FontDef const& _fontDef);
    void copyToClipboard(std::string_view _data);
    void inspect();
    void notify(std::string_view _title, std::string_view _body);
    void reply(std::string_view _response);

    template <typename... T>
    void reply(fmt::format_string<T...> fmt, T&&... args)
    {
        reply(fmt::vformat(fmt, fmt::make_format_args(args...)));
    }

    void requestWindowResize(PageSize);
    void requestWindowResize(ImageSize);
    void setApplicationkeypadMode(bool _enabled);
    void setBracketedPaste(bool _enabled);
    void setCursorStyle(CursorDisplay _display, CursorShape _shape);
    void setCursorVisibility(bool _visible);
    void setGenerateFocusEvents(bool _enabled);
    void setMouseProtocol(MouseProtocol _protocol, bool _enabled);
    void setMouseTransport(MouseTransport _transport);
    void setMouseWheelMode(InputGenerator::MouseWheelMode _mode);
    void setWindowTitle(std::string_view _title);
    [[nodiscard]] std::string const& windowTitle() const noexcept;
    void saveWindowTitle();
    void restoreWindowTitle();
    void setTerminalProfile(std::string const& _configProfileName);
    void useApplicationCursorKeys(bool _enabled);
    void softReset();
    void hardReset();
    void forceRedraw(std::function<void()> artificialSleep);
    void discardImage(Image const&);
    void markCellDirty(CellLocation _position) noexcept;
    void markRegionDirty(Rect _area) noexcept;
    void synchronizedOutput(bool _enabled);
    void onBufferScrolled(LineCount _n) noexcept;

    void setMaxImageColorRegisters(unsigned value) noexcept { state_.maxImageColorRegisters = value; }

    /// @returns either an empty string or a file:// URL of the last set working directory.
    [[nodiscard]] std::string const& currentWorkingDirectory() const noexcept
    {
        return state_.currentWorkingDirectory;
    }

    void verifyState();

    [[nodiscard]] TerminalState& state() noexcept { return state_; }
    [[nodiscard]] TerminalState const& state() const noexcept { return state_; }

    void applyPageSizeToCurrentBuffer();
    void applyPageSizeToMainDisplay(ScreenType screenType);

    [[nodiscard]] crispy::BufferObjectPtr<char> currentPtyBuffer() const noexcept
    {
        return currentPtyBuffer_;
    }

    [[nodiscard]] terminal::SelectionHelper& selectionHelper() noexcept { return selectionHelper_; }

    [[nodiscard]] Selection::OnSelectionUpdated selectionUpdatedHelper()
    {
        return [this]() {
            onSelectionUpdated();
        };
    }

    void onSelectionUpdated();

    [[nodiscard]] ViInputHandler& inputHandler() noexcept { return state_.inputHandler; }
    [[nodiscard]] ViInputHandler const& inputHandler() const noexcept { return state_.inputHandler; }
    void resetHighlight();

    StatusDisplayType statusDisplayType() const noexcept { return state_.statusDisplayType; }
    void setStatusDisplay(StatusDisplayType statusDisplayType);
    void setActiveStatusDisplay(ActiveStatusDisplay activeDisplay);

    void pushStatusDisplay(StatusDisplayType statusDisplayType);
    void popStatusDisplay();

    bool allowInput() const noexcept { return !isModeEnabled(AnsiMode::KeyboardAction); }

    void setAllowInput(bool enabled);

    // Sets the current search term to the given text and
    // moves the viewport accordingly to make sure the given text is visible,
    // or it will not move at all if the input text was not found.
    [[nodiscard]] std::optional<CellLocation> searchReverse(std::u32string text, CellLocation searchPosition);
    [[nodiscard]] std::optional<CellLocation> searchReverse(CellLocation searchPosition);

    // Searches from current position the next item downwards.
    [[nodiscard]] std::optional<CellLocation> search(std::u32string text,
                                                     CellLocation searchPosition,
                                                     bool initiatedByDoubleClick = false);
    [[nodiscard]] std::optional<CellLocation> search(CellLocation searchPosition);

    bool setNewSearchTerm(std::u32string text, bool initiatedByDoubleClick);
    void clearSearch();

    // Tests if the grid cell at the given location does contain a word delimiter.
    [[nodiscard]] bool wordDelimited(CellLocation position) const noexcept;

    [[nodiscard]] std::tuple<std::u32string, CellLocationRange> extractWordUnderCursor(
        CellLocation position) const noexcept;

    Settings const& factorySettings() const noexcept { return factorySettings_; }
    Settings const& settings() const noexcept { return settings_; }
    Settings& settings() noexcept { return settings_; }

    // Renders current visual terminal state to the render buffer.
    //
    // @param output target render buffer to write the current visual state to.
    // @param _includeSelection boolean to indicate whether or not to include colorize selection.
    void fillRenderBuffer(RenderBuffer& output, bool includeSelection); // <- acquires the lock

  private:
    void mainLoop();
    void fillRenderBufferInternal(RenderBuffer& _output, bool includeSelection);
    void updateIndicatorStatusLine();
    void updateCursorVisibilityState() const;
    bool updateCursorHoveringState();

    template <typename BlinkerState>
    [[nodiscard]] std::pair<bool, std::chrono::steady_clock::time_point> nextBlinkState(
        BlinkerState blinker, std::chrono::steady_clock::time_point lastBlink) const noexcept
    {
        auto const passed = std::chrono::duration_cast<std::chrono::milliseconds>(currentTime_ - lastBlink);
        if (passed < blinker.interval)
            return { blinker.state, lastBlink };
        return { !blinker.state, currentTime_ };
    }

    // Reads from PTY.
    [[nodiscard]] Pty::ReadResult readFromPty();

    // Writes partially or all input data to the PTY buffer object and returns a string view to it.
    [[nodiscard]] std::string_view lockedWriteToPtyBuffer(std::string_view data);

    // private data
    //

    Events& eventListener_;

    // configuration state
    Settings factorySettings_;
    Settings settings_;
    TerminalState state_;

    // synchronization
    std::mutex mutable outerLock_;
    std::mutex mutable innerLock_;

    // terminal clock
    std::chrono::steady_clock::time_point currentTime_;

    // {{{ PTY and PTY read buffer management
    crispy::BufferObjectPool<char> ptyBufferPool_;
    crispy::BufferObjectPtr<char> currentPtyBuffer_;
    size_t ptyReadBufferSize_;
    std::unique_ptr<Pty> pty_;
    // }}}

    // {{{ mouse related state (helpers for detecting double/tripple clicks)
    std::chrono::steady_clock::time_point lastClick_ {};
    unsigned int speedClicks_ = 0;
    terminal::CellLocation currentMousePosition_ {}; // current mouse position
    bool respectMouseProtocol_ = true;    // shift-click can disable that, button release sets it back to true
    bool leftMouseButtonPressed_ = false; // tracks left-mouse button pressed state (used for cell selection).
    // }}}

    // {{{ blinking state helpers
    mutable std::chrono::steady_clock::time_point lastCursorBlink_;
    mutable unsigned cursorBlinkState_;
    struct BlinkerState
    {
        bool state = false;
        std::chrono::milliseconds const interval;
    };
    mutable BlinkerState _slowBlinker { false, std::chrono::milliseconds { 500 } };
    mutable BlinkerState _rapidBlinker { false, std::chrono::milliseconds { 300 } };
    mutable std::chrono::steady_clock::time_point _lastBlink;
    mutable std::chrono::steady_clock::time_point _lastRapidBlink;
    // }}}

    // {{{ Displays this terminal manages
    // clang-format off
    Screen<PrimaryScreenCell> primaryScreen_;
    Screen<AlternateScreenCell> alternateScreen_;
    Screen<StatusDisplayCell> hostWritableStatusLineScreen_;
    Screen<StatusDisplayCell> indicatorStatusScreen_;
    std::reference_wrapper<ScreenBase> currentScreen_;
    Viewport viewport_;
    // clang-format on
    // }}}

    // {{{ selection states
    std::unique_ptr<Selection> selection_;
    struct SelectionHelper: public terminal::SelectionHelper
    {
        Terminal* terminal;
        explicit SelectionHelper(Terminal* self): terminal { self } {}
        [[nodiscard]] PageSize pageSize() const noexcept override;
        [[nodiscard]] bool wordDelimited(CellLocation _pos) const noexcept override;
        [[nodiscard]] bool wrappedLine(LineOffset _line) const noexcept override;
        [[nodiscard]] bool cellEmpty(CellLocation _pos) const noexcept override;
        [[nodiscard]] int cellWidth(CellLocation _pos) const noexcept override;
    };
    SelectionHelper selectionHelper_;
    // }}}

    // {{{ Render buffer state
    /// Boolean, indicating whether the terminal's screen buffer contains updates to be rendered.
    mutable std::atomic<uint64_t> changes_ { 0 };
    bool screenDirty_ = false; // TODO: just inc changes_ and delete this instead.
    RefreshInterval refreshInterval_;
    RenderDoubleBuffer renderBuffer_ {};
    std::atomic<uint64_t> lastFrameID_ = 0;
    RenderPassHints _lastRenderPassHints {};
    // }}}

    InputMethodData inputMethodData_ {};
    std::atomic<bool> hoveringHyperlink_ = false;
    std::atomic<bool> renderBufferUpdateEnabled_ = true; // for "Synchronized Updates" feature
    std::optional<HighlightRange> highlightRange_ = std::nullopt;
};

} // namespace terminal
